/* UDP sender to export entropy from a TRNG device.
 * J. Welsh, April & August 2021
 */

#define _BSD_SOURCE /* for gethostbyname */

#include <signal.h>
#include <stdlib.h>
#include <string.h> /* for memcpy */

#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include "io.h"

char const usage[] =
"Usage: rngsend -d TRNG_DEV [-p PORT] [--] RECEIVER ...\n"
"\n"
"Reads from a True Random Number Generator device and transmits in 512-byte UDP packets to one or more receivers in a simple rotation as the buffer is filled.\n"
"\n"
"Options:\n"
"\t-d TRNG_DEV : device path (required)\n"
"\t-p PORT : port number (default 25385)\n"
"\n"
"Positional parameters:\n"
"\tRECEIVER : IP address (or hostname, resolved at startup) for each receiver in the rotation\n";

#define PKSIZE 512 /* packet size defined by the protocol */

void load_address(struct sockaddr_in *dest, char const *name, unsigned port) {
	struct hostent *h = gethostbyname(name);
	if (!h) {
		write_err(name, hstrerror(h_errno));
		_exit(1);
	}
	if (!h->h_addr_list) {
		write_err(name, "no addresses for name");
		_exit(1);
	}
	if (h->h_addrtype != AF_INET || h->h_length != sizeof dest->sin_addr) {
		write_err("gethostbyname", "BUG: bad length or type");
		_exit(1);
	}
	dest->sin_family = AF_INET;
	dest->sin_port = htons(port);
	memcpy(&(dest->sin_addr), h->h_addr_list[0], sizeof dest->sin_addr);
}

int main(int argc_in, char *argv_in[]) {
	int trngfd, sockfd;
	struct sockaddr_in *recv_addrs;
	char buf[PKSIZE];
	unsigned port = 25385;
	char const *trng_path = 0;
	unsigned i;

	io_init("rngsend", argc_in, argv_in);
	sockfd = chkp("socket", socket(AF_INET, SOCK_DGRAM, 0));

	if (argc) pop_arg();
next_arg:
	while (!end_of_options()) {
		char const *arg = pop_arg();
		while (*++arg) {
			switch (*arg) {
				case 'd':
					++arg;
					if (!*arg) {
						if (!argc) usage_err("-d missing value");
						arg = pop_arg();
					}
					trng_path = arg;
					goto next_arg;
				case 'h':
					write_str(1, usage);
					return 0;
				case 'p':
					++arg;
					if (!*arg) {
						if (!argc) usage_err("-p missing value");
						arg = pop_arg();
					}
					port = parse_uint_option(arg);
					if (port > 65535) usage_err("port number out of range");
					goto next_arg;
				default:
					usage_err("bad option");
			}
		}
	}

	if (!trng_path) usage_err("missing TRNG device path");
	trngfd = chkp(trng_path, open(trng_path, O_RDONLY));

	if (!argc) usage_err("missing receiver address list");
	recv_addrs = calloc(argc, sizeof *recv_addrs);
	if (!recv_addrs) {
		perr("calloc");
		_exit(1);
	}
	for (i=0; i<argc; ++i) {
		load_address(&recv_addrs[i], argv[i], port);
	}

	write_line(1, "sender started");
	for (;;) {
		/* For each client in rotation, fill the buffer and send it off. If a packet is lost, no big deal, another will follow soon enough. With a true entropy source (which is the whole point) this is not a busy loop because the read blocks. */
		for (i=0; i<argc; ++i) {
			if (read_all(trngfd, buf, sizeof buf) < sizeof buf) {
				write_err(trng_path, "EOF on read from TRNG device");
				_exit(1);
			}
			if (sendto(sockfd, buf, sizeof buf, 0,
						(struct sockaddr *)(&recv_addrs[i]),
						sizeof *recv_addrs) == -1) {
				/* This might not be possible, at least on Linux. Log the error but don't die, on the theory that it's temporary. */
				perr("sendto");
			}
		}
	}

	return 0;
}
